import unittest

from ..session import get_cached_retry_session, get_retry_session


class LiveSessionTest(unittest.TestCase):
    def test_just_call_get_retry_session(self) -> None:
        session = get_retry_session()
        session.get("https://httpbin.org/status/200,503")

    def test_just_call_get_cached_retry_session(self) -> None:
        session = get_cached_retry_session()
        session.get("https://httpbin.org/status/200,503")



import logging
import os
import threading
import time
from pathlib import Path
from typing import Dict, IO, List, Optional

import click
import keke

from indexurl import get_index_url
from indexurl.core import DEFAULT_INDEX_URL
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from pypi_simple import ACCEPT_JSON_PREFERRED, PyPISimple
from vmodule import vmodule_init

from .cache import NoCache, SimpleCache
from .markers import EnvironmentMarkers
from .resolution import Walker
from .session import get_cached_retry_session, get_retry_session
from .types import CanonicalName

LOG = logging.getLogger(__name__)


def _stats_thread() -> None:
    prev_ts = None
    prev_process_time = None
    while True:
        ts = time.time()
        process_time = time.process_time()
        if prev_ts is not None:
            keke.kcount(
                "proc_cpu_pct",
                100 * (process_time - prev_process_time) / (ts - prev_ts),
            )

        prev_ts = ts
        prev_process_time = process_time
        time.sleep(0.1)


@click.command()
@click.pass_context
@click.option(
    "--trace", type=click.File("w"), help="Write chrome trace to this filename"
)
@click.option("--stats", is_flag=True, help="Include cpu stats in the trace")
@click.option(
    "--verbose",
    "-v",
    type=int,
    help="Enable verbose logging (default=WARNING, 0=INFO, 1=VLOG_1, 10=DEBUG)",
)
@click.option(
    "--vmodule",
    help="Enable verbose logging only for some explicitly named loggers (e.g. "
    "'hdeps.resolution=1' would enable VLOG_1 for just that one logger.  Comma-separated.",
)
@click.option(
    "-I",
    "--isolate-env",
    is_flag=True,
    help="Isolate from the user's environment (use default index url).",
)
@click.option(
    "--no-cache",
    is_flag=True,
    help="Do not read or write cache for dependencies.",
)
@click.option(
    "--parallelism",
    "-p",
    default=24,
    type=int,
    help="Parallelism factor for network i/o",
)
@click.option(
    "--platform",
    metavar="PLAT",
    default="linux",
    show_default=True,
    help="Optionally override platform: linux, win32, darwin.",
)
@click.option(
    "--python-version",
    metavar="VERSION",
    help="Optionally override python version.  Default is autodetect running.",
)
@click.option(
    "--install-order",
    is_flag=True,
    help="Output a theoretical install order instead of a tree",
)
@click.option(
    "--print-legend",
    is_flag=True,
    help="Output the meaning of colors in a header",
)
@click.option(
    "--color/--no-color",
    is_flag=True,
    default=None,
    help="Default is to guess from NO_COLOR or FORCE_COLOR env vars being non-empty",
)
@click.option("--have", help="pkg==ver to assume already installed", multiple=True)
@click.option("-r", "--requirements-file", multiple=True)
@click.argument(
    "deps",
    nargs=-1,
)
def main(
    ctx: click.Context,
    trace: Optional[IO[str]],
    stats: bool,
    verbose: Optional[int],
    vmodule: Optional[str],
    parallelism: int,
    have: List[str],
    requirements_file: List[str],
    deps: List[str],
    platform: Optional[str],
    python_version: Optional[str],
    install_order: bool,
    isolate_env: bool,
    no_cache: bool,
    print_legend: bool,
    color: Optional[bool],
) -> None:
    if trace:
        ctx.with_resource(keke.TraceOutput(trace))
    vmodule_init(verbose, vmodule)
    if stats:
        threading.Thread(target=_stats_thread, daemon=True).start()

    uncached_session = get_retry_session()
    extracted_metadata_cache: SimpleCache
    if no_cache:
        cached_session = uncached_session
        extracted_metadata_cache = NoCache()
    else:
        cached_session = get_cached_retry_session()
        extracted_metadata_cache = SimpleCache()

    if isolate_env:
        index_url = DEFAULT_INDEX_URL
    else:
        index_url = get_index_url()

    if color is True or (color is None and os.environ.get("FORCE_COLOR")):
        ctx.color = True
    elif color is False or (color is None and os.environ.get("NO_COLOR")):
        ctx.color = False

    have_versions: Dict[CanonicalName, str] = {}
    for h in have:
        k, op, v = h.partition("==")
        if op != "==":
            raise click.ClickException(f"Invalid format for --have: {h!r}")
        have_versions[CanonicalName(canonicalize_name(k))] = v

    walker = Walker(
        parallelism,
        env_markers=EnvironmentMarkers.from_args(
            python_version=python_version, sys_platform=platform
        ),
        pypi_simple=PyPISimple(
            index_url, session=cached_session, accept=ACCEPT_JSON_PREFERRED
        ),
        uncached_session=uncached_session,
        current_version_callback=have_versions.get,
        extracted_metadata_cache=extracted_metadata_cache,
        color=ctx.color,
    )

    def solve() -> None:
        for dep in deps:
            walker.feed(Requirement(dep))
        for req in requirements_file:
            walker.feed_file(Path(req))
        walker.drain()

    solve()

    if install_order:
        walker.print_flat()
    else:
        if print_legend:
            walker.print_legend()
        walker.print_tree()

    click.echo("========== Summary ==========")
    if walker.known_conflicts:
        resolutions: List[Requirement] = []
        for project, versions in walker.known_conflicts.copy().items():
            click.echo(
                f"Found conflict: {project} {sorted([str(x) for x in versions])}"
            )
            for version in versions:
                with keke.kev("pin_attempt", project=project, version=version):
                    LOG.info("Trying to pin %s==%s", project, version)
                    req = Requirement(f"{project}=={version}")
                    walker.clear()
                    # Add the pin as first requirement to see if it can be used
                    # by all subsequent dependencies.
                    walker.feed(req)
                    solve()
                if project not in walker.known_conflicts:
                    resolutions.append(req)
                    break

        if resolutions:
            click.echo("Pin these project versions to resolve conflict:")
            for resolution in resolutions:
                click.echo(resolution)

        unresolved = walker.known_conflicts.keys() - {x.name for x in resolutions}
        if unresolved:
            click.echo("Failed to resolve following conflicts:")
            for conflict in sorted(unresolved):
                click.echo(
                    f"{conflict} {sorted([str(x) for x in walker.known_conflicts[conflict]])}"
                )
    else:
        click.echo("No conflicts found.")


if __name__ == "__main__":
    main()



import io
from pathlib import Path
from typing import Any

from requests import Response, Session


class FakeSession:
    def __init__(self, fixture_root: Path) -> None:
        self.fixture_root = fixture_root

    def request(self, method: str, *args: Any, **kwargs: Any) -> Response:
        if method.lower() == "get":
            return self.get(*args, **kwargs)
        else:
            raise NotImplementedError(method)

    def get(
        self, url: str, headers: Any = {}, timeout: float = 0, stream: bool = False
    ) -> Response:
        # This is intended to "serve" any files from self.fixture_root, but git
        # checkouts on windows with core.autocrlf alter the line endings of
        # files it thinks are text.  We need the checksums of some text files
        # like the mime documents *.metadata to be consistent.
        if headers is None:
            headers = {}
        text = True
        if url.endswith(".metadata"):
            parts = url.split("/")
            local_path = self.fixture_root / parts[-1]
        elif url.endswith((".gz", ".zip", ".whl")):
            parts = url.split("/")
            local_path = self.fixture_root / parts[-1]
            text = False
        elif "/simple/" in url:
            project = url.strip("/").split("/")[-1]
            local_path = self.fixture_root / f"{project}.html"
        else:
            raise ValueError(f"Unhandled path {url}")

        if not local_path.exists():
            resp = Response()
            resp.status_code = 404
            return resp

        data = local_path.read_bytes()
        if text:
            data = data.replace(b"\r\n", b"\n")

        length = str(len(data))

        resp = Response()
        resp.url = url
        resp.status_code = 200

        if range := headers.get("Range"):
            assert not text
            resp.status_code = 206
            x = range.rsplit("=", 1)[-1]
            if x[0] == "-":
                # bytes from end
                start = max(0, len(data) + int(x))
                data = data[int(x) :]
            else:
                a, b = x.split("-", 1)
                start = int(a)
                data = data[int(a) : int(b) + 1]
            # TODO make these better numbers
            resp.headers["content-range"] = f"bytes {start}-1/{length}"

        resp.raw = io.BytesIO(data)
        resp.headers["content-type"] = "text/html"
        resp.headers["content-length"] = str(len(data))
        return resp


def get_fake_session_fixtures() -> Session:
    return FakeSession(Path(__file__).parent / "fixtures")  # type: ignore



import unittest

from packaging.requirements import Requirement

from ..markers import EnvironmentMarkers


class EnvironmentMarkersTest(unittest.TestCase):
    def test_platforms(self) -> None:
        e = EnvironmentMarkers(sys_platform="win32")
        self.assertEqual("nt", e.os_name)
        e = EnvironmentMarkers(sys_platform="darwin")
        self.assertEqual("posix", e.os_name)
        e = EnvironmentMarkers(python_version="2.7.5")
        self.assertEqual("linux2", e.sys_platform)
        with self.assertRaises(TypeError):
            e = EnvironmentMarkers(sys_platform="x")

    def test_match(self) -> None:
        req = Requirement("foo ; python_version == '3.6'")

        e = EnvironmentMarkers(python_version="3.6")
        self.assertTrue(e.match(req.marker))

        e = EnvironmentMarkers(python_version="3.7")
        self.assertFalse(e.match(req.marker))

    def test_extras_match(self) -> None:
        req = Requirement("foo ; python_version == '3.6' and extra == 'x'")

        e = EnvironmentMarkers(python_version="3.6")
        self.assertFalse(e.match(req.marker))
        self.assertFalse(e.match(req.marker, []))
        self.assertTrue(e.match(req.marker, ["x"]))
        self.assertTrue(e.match(req.marker, ["x", "y"]))

        e = EnvironmentMarkers(python_version="3.7")
        self.assertFalse(e.match(req.marker))
        self.assertFalse(e.match(req.marker, []))
        self.assertFalse(e.match(req.marker, ["x"]))
        self.assertFalse(e.match(req.marker, ["x", "y"]))



import sys
from dataclasses import asdict, dataclass, replace
from typing import Optional, Sequence

from packaging.markers import Marker


@dataclass
class EnvironmentMarkers:
    os_name: str = "posix"
    sys_platform: str = "linux"
    platform_machine: str = "x86_64"
    platform_python_implementation: str = "CPython"
    platform_release: Optional[str] = None
    platform_system: str = "Linux"
    platform_version: Optional[str] = None
    python_version: Optional[str] = None
    python_full_version: Optional[str] = None
    implementation_name: str = "cpython"

    def __post_init__(self) -> None:
        if self.sys_platform == "linux":
            if self.python_version and self.python_version[:1] == "2":
                self.sys_platform = "linux2"
        elif self.sys_platform == "win32":
            self.platform_system = "Windows"
            self.os_name = "nt"
        elif self.sys_platform == "darwin":
            self.platform_system = "Darwin"
        else:
            raise TypeError(f"Unknown sys_platform: {self.sys_platform!r}")

    def match(self, marker: Optional[Marker], extras: Sequence[str] = ()) -> bool:
        env = asdict(self)
        if marker:
            if extras:
                if not any(marker.evaluate(dict(env, extra=e)) for e in extras):
                    return False
            else:
                if not marker.evaluate(env):
                    return False
        return True

    @classmethod
    def from_args(
        cls, python_version: Optional[str], sys_platform: Optional[str]
    ) -> "EnvironmentMarkers":
        if python_version:
            if python_version.count(".") == 1:
                python_version += ".0"
        else:
            python_version = ".".join(str(v) for v in sys.version_info[:3])

        obj = cls(
            python_version=python_version.rsplit(".", 1)[0],
            python_full_version=python_version,
        )
        if sys_platform:
            obj = replace(obj, sys_platform=sys_platform)
        return obj



import os
import tempfile
import unittest
from pathlib import Path

from ..requirements import iter_glob_all_requirement_names, iter_requirement_names


class RequirementsTest(unittest.TestCase):
    def test_iter_requirement_names(self) -> None:
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(
                b"""\
--index-url foo
a

# comment
B # inline comment
c==2
"""
            )
            f.close()
            self.assertEqual(
                ["a", "b", "c"], list(iter_requirement_names(Path(f.name)))
            )

    def test_iter_glob_all_requirement_names(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)

            (pd / "a").mkdir()
            (pd / "test").mkdir()
            (pd / "tbat").mkdir()

            (pd / "requirements.txt").write_text("x\n")
            (pd / "a" / "requirements.txt").write_text("a==1\n")
            (pd / "test" / "requirements.txt").write_text("b==1\n")
            (pd / "tbat" / "requirements.txt").write_text("c==1\n")
            prev = os.getcwd()
            try:
                os.chdir(d)
                self.assertEqual(
                    ["b", "c", "x"],
                    sorted(
                        iter_glob_all_requirement_names("*.txt,t*/requirements.txt,")
                    ),
                )
            finally:
                os.chdir(prev)



import logging
import threading
from collections import defaultdict, deque
from concurrent.futures import Future, ThreadPoolExecutor

from pathlib import Path
from typing import Any, Dict, Optional, Set, Tuple

import click

from keke import kev, ktrace

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version
from pypi_simple import NoSuchProjectError, PyPISimple
from requests.sessions import Session
from vmodule import VLOG_1, VLOG_2

# from tpe_prio import ThreadPoolExecutor

from .cache import SimpleCache

from .compatibility import find_best_compatible_version, NoMatchingRelease
from .markers import EnvironmentMarkers
from .projects import BasicMetadata, Project, ProjectVersion
from .requirements import _iter_simple_requirements
from .session import get_retry_session
from .types import CanonicalName, Choice, ChoiceKeyType, Edge, VersionCallback

LOG = logging.getLogger(__name__)


def _all_current_versions_unknown(cn: CanonicalName) -> Optional[str]:
    """
    Default for when you don't need to reuse any already-installed project
    versions.
    """
    return None


class Walker:
    def __init__(
        self,
        parallelism: int,
        env_markers: EnvironmentMarkers,
        pypi_simple: PyPISimple,
        uncached_session: Optional[Session] = None,
        current_version_callback: VersionCallback = _all_current_versions_unknown,
        extracted_metadata_cache: Optional[SimpleCache] = None,
        color: Optional[bool] = None,
    ):
        self.root = Choice(CanonicalName("-"), Version("0"))
        self.pool = ThreadPoolExecutor(max_workers=parallelism)
        self.env_markers = env_markers
        self.pypi_simple = pypi_simple
        self.uncached_session = uncached_session or get_retry_session()
        self.extracted_metadata_cache = extracted_metadata_cache or SimpleCache()
        self.color = color

        self.memo_fetch: Dict[CanonicalName, Future[Optional[Project]]] = {}
        self.memo_fetch_lock = threading.Lock()
        self.memo_version_metadata: Dict[ProjectVersion, Future[BasicMetadata]] = {}

        self.queue: deque[
            Tuple[Choice, CanonicalName, Requirement, str, Set[ChoiceKeyType]]
        ] = deque()
        self.current_version_callback = current_version_callback
        self.known_conflicts: Dict[CanonicalName, Set[Version]] = defaultdict(set)

    def clear(self) -> None:
        self.root = Choice(CanonicalName("-"), Version("0"))
        self.known_conflicts.clear()

    def feed_file(self, req_file: Path) -> None:
        for req in _iter_simple_requirements(req_file):
            self.feed(req, str(req_file))

    def feed(self, req: Requirement, source: str = "arg") -> None:
        name = CanonicalName(canonicalize_name(req.name))
        LOG.log(VLOG_1, "Feed %s (%r) from %s", name, str(req), source)
        if req.marker and not self.env_markers.match(req.marker):
            return

        with self.memo_fetch_lock:
            if name not in self.memo_fetch:
                self.memo_fetch[name] = self.pool.submit(
                    self._fetch_project, name, False
                )

        empty_set: Set[ChoiceKeyType] = set()
        self.queue.append((self.root, name, req, source, empty_set))

    @ktrace("project_name", "proactive", shortname=True)
    def _fetch_project(
        self, project_name: CanonicalName, proactive: bool
    ) -> Optional[Project]:
        try:
            project_page = self.pypi_simple.get_project_page(project_name)
        except NoSuchProjectError:
            return None
        project = Project.from_pypi_simple_project_page(project_page)
        # It's extremely likely that we will subsequently look up the deps of
        # the most recent version, so go ahead and schedule the metadata fetch.
        if project.versions:
            latest_version_key = max(project.versions)
            latest_version = project.versions[latest_version_key]

            if latest_version not in self.memo_version_metadata:
                self.memo_version_metadata[latest_version] = self.pool.submit(
                    self._fetch_project_metadata, project_name, latest_version
                )

        LOG.log(VLOG_1, "_fetch_project done %s", project_name)
        return project

    @ktrace("project_name", "version.version", shortname=True)
    def _fetch_project_metadata(
        self, project_name: CanonicalName, version: ProjectVersion
    ) -> BasicMetadata:
        md = version.get_deps(
            self.pypi_simple, self.uncached_session, self.extracted_metadata_cache
        )
        # It's extremely likely that we will subsequently look up the deps of
        # this version, so go ahead and schedule that too.  (But not with any
        # extras.)
        with kev("prefetch", count=len(md.reqs)):
            for req in md.reqs:
                name = CanonicalName(canonicalize_name(req.name))
                # Don't bother with markers if we've already scheduled
                if name not in self.memo_fetch:
                    # Marker evaluation is relatively expensive
                    if self.env_markers.match(req.marker):
                        # Check again under the lock (for correctness, and since some time has elapsed)
                        with self.memo_fetch_lock:
                            if name not in self.memo_fetch:
                                self.memo_fetch[name] = self.pool.submit(
                                    self._fetch_project, name, True
                                )

        LOG.log(
            VLOG_1, "_fetch_project_metadata done %s %s", project_name, version.version
        )
        return md

    def drain(self) -> None:
        chosen: Dict[CanonicalName, Version] = {}

        while self.queue:
            (parent, name, req, source, parent_keys) = self.queue.popleft()
            LOG.info(
                "process %s %s from %s with extras %s", name, req, source, req.extras
            )
            fut = self.memo_fetch[name]
            if hasattr(self.pool, "bump"):
                self.pool.bump(fut)
            with kev("project result", project_name=name):
                project = fut.result()

            if project is None:
                LOG.error("Missing dep processing %s", name)
                continue

            cur = chosen.get(name)
            with kev("find_best_compatible_version", project_name=name, req=str(req)):
                try:
                    version = find_best_compatible_version(
                        project,
                        req,
                        self.env_markers,
                        cur,
                        self.current_version_callback,
                    )
                except NoMatchingRelease:
                    LOG.error("No matching version for %s processing %s", req, name)
                    continue
            choice = Choice(name, version)

            edge = Edge(
                choice, specifier=req.specifier, markers=req.marker, note=source
            )
            parent.deps.append(edge)
            if choice.key() in parent_keys:
                LOG.info("Avoid circular dep processing %s", name)
                continue

            if cur and cur != version:
                LOG.info("Multiple versions for %s: %s and %s", name, cur, version)
                self.known_conflicts[name].update([cur, version])
            chosen[name] = version

            if t := project.versions.get(version):
                if t not in self.memo_version_metadata:
                    fut2 = self.pool.submit(self._fetch_project_metadata, name, t)
                    self.memo_version_metadata[t] = fut2
                else:
                    fut2 = self.memo_version_metadata[t]

                if hasattr(self.pool, "bump"):
                    self.pool.bump(fut2)
                with kev("ver result", project_name=name, project_version=str(version)):
                    md = fut2.result()

                choice.has_sdist = md.has_sdist
                choice.has_wheel = md.has_wheel
                for r in md.reqs:
                    r_name = CanonicalName(canonicalize_name(r.name))
                    LOG.log(
                        VLOG_2,
                        "  drain possible requirement %s -> %s (%s)",
                        name,
                        r_name,
                        r,
                    )
                    if self.env_markers.match(r.marker, sorted(req.extras)):
                        with self.memo_fetch_lock:
                            if r_name not in self.memo_fetch:
                                fut = self.pool.submit(
                                    self._fetch_project, r_name, False
                                )
                                self.memo_fetch[r_name] = fut

                        self.queue.append(
                            (choice, r_name, r, "dep", parent_keys | {choice.key()})
                        )

                        LOG.log(VLOG_2, "    keep")
                    else:
                        LOG.log(VLOG_2, "    omit")

    def print_flat(
        self,
        choice: Optional[Choice] = None,
        seen: Optional[
            Set[Tuple[CanonicalName, Version, Optional[Tuple[str, ...]]]]
        ] = None,
    ) -> None:
        if choice is None:
            choice = self.root
            seen = set()

        assert seen is not None
        # Simple recursive postorder
        for x in choice.deps:
            key = (x.target.project, x.target.version, x.target.extras)
            flag = key in seen
            seen.add(key)

            if x.target.deps:
                self.print_flat(x.target, seen)
            dep_extras = f"[{', '.join(x.target.extras)}]" if x.target.extras else ""
            if not flag:
                print(f"{x.target.project}{dep_extras}=={x.target.version}")

    COLORS: Dict[Optional[str], Optional[str]] = {
        "conflict": "magenta",
        "good": "green",
        "have_reuse": "cyan",
        "no_sdist": "red",
        "no_wheel": "blue",
        None: None,
    }

    def print_legend(self) -> None:
        click.echo(
            click.style("[good]", fg=self.COLORS["good"]) + " is what you hope to see."
        )
        click.echo(
            click.style("[conflict]", fg=self.COLORS["conflict"])
            + " means two different versions were found during this walk."
        )
        click.echo(
            click.style("[no_sdist]", fg=self.COLORS["no_sdist"])
            + " means this project does not have an sdist.  (This is something"
            + " to watch out for if you want to build from source.)"
        )
        # This is not a whole-line styling -- omit for now
        # click.echo(
        #     click.style("[no_wheel]", fg=self.COLORS["no_wheel"])
        #     + " means this project does not have a wheel, thus might be missing its deps."
        # )
        click.echo(
            click.style("[have_reuse]", fg=self.COLORS["have_reuse"])
            + " means that a version specified in --have was kept."
        )
        click.echo()

    def print_tree(
        self,
        choice: Optional[Choice] = None,
        seen: Optional[
            Set[Tuple[CanonicalName, Version, Optional[Tuple[str, ...]]]]
        ] = None,
        known_conflicts: Dict[CanonicalName, Set[Version]] = defaultdict(set),
        depth: int = 0,
    ) -> None:
        prefix = ". " * depth
        if choice is None:
            choice = self.root
            seen = set()
            known_conflicts = self.known_conflicts

        assert seen is not None
        # Inorder, but avoid doing duplicate work...
        for x in choice.deps:
            # TODO display whether install or build dep, and whether pin disallows
            # current version, has compatible bdist, no sdist, etc
            key = (x.target.project, x.target.version, x.target.extras)
            dep_extras = (
                f"[{', '.join(sorted(x.target.extras))}]" if x.target.extras else ""
            )

            def print_line(color_choice: Optional[str], dep_stuff: Any) -> None:
                click.echo(
                    prefix
                    + click.style(x.target.project, fg=self.COLORS[color_choice])
                    + dep_stuff
                    + (
                        f" [{color_choice}]"
                        # N.b. self.color is intentionally tri-state -- None
                        # being autodetect in click and we assume it will be
                        # enabled thus don't output names here.
                        if color_choice and self.color is False
                        else ""
                    )
                )

            if key in seen:
                print_line(
                    "conflict" if key[0] in known_conflicts and x.specifier else None,
                    f"{dep_extras} (=={x.target.version}) (already listed){' ; ' + str(x.markers) if x.markers else ''} via "
                    + click.style(x.specifier or "*", fg="yellow"),
                )
            else:
                if key[0] in known_conflicts:
                    # conflicting decision
                    color = "conflict"
                else:
                    cur = self.current_version_callback(x.target.project)
                    if cur and Version(cur) == x.target.version:
                        color = "have_reuse"
                    else:
                        color = "no_sdist" if not x.target.has_sdist else "good"
                seen.add(key)

                print_line(
                    color,
                    f"{dep_extras} (=={x.target.version}){' ; ' + str(x.markers) if x.markers else ''} via "
                    + click.style(x.specifier or "*", fg="yellow")
                    + click.style(
                        " no whl" if not x.target.has_wheel else "", fg="blue"
                    ),
                )
                if x.target.deps:
                    self.print_tree(x.target, seen, known_conflicts, depth + 1)



import tempfile
import unittest
from pathlib import Path

from packaging.requirements import Requirement
from packaging.version import Version

from pypi_simple import DistributionPackage, PyPISimple

from ..cache import SimpleCache
from ..projects import ProjectVersion

from ._fake_session import FakeSession

DIST_ROOT = Path(__file__).parent.joinpath("demo_project/dist")

EXPECTED_REQS = [
    Requirement("a"),
    Requirement('b; python_version == "3.6"'),
    Requirement('c; extra == "foo"'),
    Requirement('d; python_version == "3.6" and extra == "foo"'),
]


class ProjectMetadataTest(unittest.TestCase):
    def test_wheel(self) -> None:
        pv = ProjectVersion(
            Version("0.0.0"),
            (
                DistributionPackage(
                    project="demo_project",
                    version="0.0.0",
                    digests={},
                    requires_python=None,
                    has_sig=False,
                    filename="demo_project-0.0.0-py3-none-any.whl",
                    url="demo_project-0.0.0-py3-none-any.whl",
                    package_type="wheel",
                ),
            ),
        )
        with tempfile.TemporaryDirectory() as d:
            cache = SimpleCache(Path(d))
            md = pv.get_deps(
                ps=None,  # type: ignore
                session=FakeSession(DIST_ROOT),  # type: ignore
                extracted_metadata_cache=cache,
            )
            self.assertEqual(EXPECTED_REQS, md.reqs)
            self.assertEqual(["foo"], md.extras)
            self.assertEqual(1, cache.stats["pass"])
            self.assertEqual(1, cache.stats["sets"])

            # Now load from cache
            md = pv.get_deps(
                ps=None,  # type: ignore
                session=FakeSession(DIST_ROOT),  # type: ignore
                extracted_metadata_cache=cache,
            )
            self.assertEqual(EXPECTED_REQS, md.reqs)
            self.assertEqual(["foo"], md.extras)
            self.assertEqual(1, cache.stats["hits"])  # inc just this
            self.assertEqual(1, cache.stats["pass"])
            self.assertEqual(1, cache.stats["sets"])

    def test_zip_sdist(self) -> None:
        pv = ProjectVersion(
            Version("0.0.0"),
            (
                DistributionPackage(
                    project="demo_project",
                    version="0.0.0",
                    digests={},
                    requires_python=None,
                    has_sig=False,
                    filename="demo_project-0.0.0.zip",
                    url="demo_project-0.0.0.zip",
                    package_type="sdist",
                ),
            ),
        )
        with tempfile.TemporaryDirectory() as d:
            cache = SimpleCache(Path(d))
            md = pv.get_deps(
                ps=None,  # type: ignore
                session=FakeSession(DIST_ROOT),  # type: ignore
                extracted_metadata_cache=cache,
            )
            self.assertEqual(EXPECTED_REQS, md.reqs)
            self.assertEqual(["foo"], md.extras)
            self.assertEqual(1, cache.stats["pass"])
            self.assertEqual(1, cache.stats["sets"])

            # Now load from cache
            md = pv.get_deps(
                ps=None,  # type: ignore
                session=FakeSession(DIST_ROOT),  # type: ignore
                extracted_metadata_cache=cache,
            )
            self.assertEqual(EXPECTED_REQS, md.reqs)
            self.assertEqual(["foo"], md.extras)
            self.assertEqual(1, cache.stats["hits"])  # inc just this
            self.assertEqual(1, cache.stats["pass"])
            self.assertEqual(1, cache.stats["sets"])

    def test_tar_gz_sdist(self) -> None:
        pv = ProjectVersion(
            Version("0.0.0"),
            (
                DistributionPackage(
                    project="demo_project",
                    version="0.0.0",
                    digests={},
                    requires_python=None,
                    has_sig=False,
                    filename="demo_project-0.0.0.tar.gz",
                    url="demo_project-0.0.0.tar.gz",
                    package_type="sdist",
                ),
            ),
        )
        with tempfile.TemporaryDirectory() as d:
            cache = SimpleCache(Path(d))
            md = pv.get_deps(
                ps=PyPISimple(session=FakeSession(DIST_ROOT)),  # type: ignore
                session=None,  # type: ignore
                extracted_metadata_cache=cache,
            )
            self.assertEqual(EXPECTED_REQS, md.reqs)
            self.assertEqual(["foo"], md.extras)
            self.assertEqual(1, cache.stats["pass"])
            self.assertEqual(1, cache.stats["sets"])

            # Now load from cache
            md = pv.get_deps(
                ps=PyPISimple(session=FakeSession(DIST_ROOT)),  # type: ignore
                session=None,  # type: ignore
                extracted_metadata_cache=cache,
            )
            self.assertEqual(EXPECTED_REQS, md.reqs)
            self.assertEqual(["foo"], md.extras)

            self.assertEqual(1, cache.stats["hits"])  # inc just this
            self.assertEqual(1, cache.stats["pass"])
            self.assertEqual(1, cache.stats["sets"])



import unittest
from typing import Dict
from unittest.mock import Mock

from packaging.version import Version

from ..compatibility import requires_python_match


class CompatibilityTest(unittest.TestCase):
    def test_requires_python_match(self) -> None:
        # pretend the env marker says 3.7
        python_version = Version("3.7")
        cache: Dict[str, bool] = {}
        fake_project = Mock(
            versions={
                Version("3.6"): Mock(requires_python="zzz"),
                Version("3.7"): Mock(requires_python=">=3.7"),
                Version("3.8"): Mock(requires_python=">=3.8"),
            }
        )
        # This is repeated to test that we read the cached value correctly.
        self.assertFalse(
            requires_python_match(fake_project, cache, python_version, Version("3.8"))
        )
        self.assertEqual(False, cache[">=3.8"])
        self.assertFalse(
            requires_python_match(fake_project, cache, python_version, Version("3.8"))
        )
        self.assertEqual(False, cache[">=3.8"])

        # This too
        self.assertTrue(
            requires_python_match(fake_project, cache, python_version, Version("3.7"))
        )
        self.assertEqual(True, cache[">=3.7"])
        self.assertTrue(
            requires_python_match(fake_project, cache, python_version, Version("3.7"))
        )
        self.assertEqual(True, cache[">=3.7"])

        # This too
        self.assertTrue(
            requires_python_match(fake_project, cache, python_version, Version("3.6"))
        )
        self.assertEqual(True, cache["zzz"])
        self.assertTrue(
            requires_python_match(fake_project, cache, python_version, Version("3.6"))
        )
        self.assertEqual(True, cache["zzz"])



from typing import Optional

import appdirs
from cachecontrol import CacheControlAdapter
from cachecontrol.caches import SeparateBodyFileCache
from requests.adapters import HTTPAdapter, Retry
from requests.sessions import Session


def get_cached_retry_session(cache_dir: Optional[str] = None) -> Session:
    if not cache_dir:
        cache_dir = appdirs.user_cache_dir("hdeps", "python-packaging")

    sess = Session()
    # Settings copied from pip/_internal/network/session.py
    retries = Retry(
        total=3, backoff_factor=0.25, status_forcelist=[500, 502, 503, 520, 527]
    )
    assert cache_dir is not None
    cache_adapter = CacheControlAdapter(
        cache=SeparateBodyFileCache(cache_dir), max_retries=retries, pool_maxsize=100
    )
    sess.mount("https://", cache_adapter)
    sess.mount("http://", cache_adapter)
    return sess


def get_retry_session() -> Session:
    sess = Session()
    # Settings copied from pip/_internal/network/session.py
    retries = Retry(
        total=3, backoff_factor=0.25, status_forcelist=[500, 502, 503, 520, 527]
    )
    adapter = HTTPAdapter(max_retries=retries, pool_maxsize=100)
    sess.mount("https://", adapter)
    sess.mount("http://", adapter)
    return sess


if __name__ == "__main__":
    with get_cached_retry_session() as s:
        s.get(
            "https://pypi.org/simple/example-foo/",
            headers={"cache-control": "max-age=0"},
            # , "range": "bytes=1-10"},
        )



import logging
import os
import re
import shlex
import unittest
from pathlib import Path
from typing import List, Optional, Tuple
from unittest.mock import patch

from click.testing import CliRunner
from parameterized import parameterized

from ..cache import NoCache
from ..cli import main

from ._fake_session import get_fake_session_fixtures

SCENARIOS: List[Tuple[str, Path]] = sorted(
    # The first item in the tuple is special for parameterized and gets mangled
    # into the test name.  It needs to be a valid identifier, so we still use
    # subTest below to get the actual filename printed for copy-pasting.
    (p.with_suffix("").name, p)
    for p in Path(__file__).parent.joinpath("scenarios").glob("*.txt")
)

LOG_LINE_TIMESTAMP_RE = re.compile(r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3} ", re.M)
LOG_LINE_NUMERIC_LINE_RE = re.compile(r"^([A-Z]+\s+[a-z_.]+:)\d+(?= )", re.M)


def load_scenario(path: Path) -> Tuple[Tuple[str, ...], str]:
    command: Optional[Tuple[str, ...]] = None
    output: str = ""
    state = 0
    with open(path) as f:
        for line in f:
            if state == 0 and line.startswith("#"):
                pass
            elif state == 0 and line.startswith("$"):
                parts = shlex.split(line[1:])
                assert parts[0] == "hdeps"
                command = tuple(parts[1:])
                state = 1
            elif state == 1:
                output += line

    assert state == 1
    assert command is not None
    return (command, output)


def save_scenario(path: Path, new_output: str) -> None:
    state = 0
    buf = ""
    with open(path) as f:
        for line in f:
            if state == 0 and line.startswith("#"):
                buf += line
            elif state == 0 and line.startswith("$"):
                buf += line
                state = 1

    assert state == 1
    if not buf.endswith("\n"):
        buf += "\n"
    path.write_text(buf + new_output)


class CliScenariosTest(unittest.TestCase):
    maxDiff = None

    @parameterized.expand(SCENARIOS)  # type:ignore[misc]
    @patch("hdeps.cli.get_retry_session", get_fake_session_fixtures)
    @patch("hdeps.cli.get_cached_retry_session", get_fake_session_fixtures)
    @patch("hdeps.cli.SimpleCache", lambda: NoCache())
    def test_scenario(self, _unused_name: str, path: Path) -> None:
        with self.subTest(path):
            command, output = load_scenario(path)

            runner = CliRunner()
            with runner.isolated_filesystem():
                del logging.root.handlers[:]
                result = runner.invoke(main, command, catch_exceptions=False)

            cleaned_output = LOG_LINE_TIMESTAMP_RE.sub("", result.output)
            cleaned_output = LOG_LINE_NUMERIC_LINE_RE.sub(
                lambda m: (m.group(1) + "<n>"), cleaned_output
            )

            if os.getenv("UPDATE_SCENARIOS"):
                save_scenario(path, cleaned_output)
            else:
                self.assertEqual(output, cleaned_output)



# This is an especially simple version of requirements parsing; we do the simple
# thing here to avoid extra deps or fragile APIs, at the expense of missing some
# deps and false-positives.

from glob import glob
from pathlib import Path
from typing import Iterator

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

# These all have iter- prefixes because I expect a more public api to pick a
# couple and return sets instead.


def _iter_simple_requirements(path: Path) -> Iterator[Requirement]:
    for line in path.read_text().splitlines():
        line = line.split("#", 1)[0].strip()
        if not line:
            continue
        if line.startswith("-"):
            print("Ignoring", line)
            continue

        # N.b. Requirement does not canonicalize its name
        yield Requirement(line)


def iter_requirement_names(path: Path) -> Iterator[str]:
    """
    Returns the canonical names from the given requirements.txt
    """
    # TODO support, or document non-support, for git references

    for req in _iter_simple_requirements(path):
        yield canonicalize_name(req.name)


def iter_glob_all_requirement_names(comma_separated_patterns: str) -> Iterator[str]:
    for pattern in comma_separated_patterns.split(","):
        if pattern:
            for filename in sorted(glob(pattern)):
                # We can't just use Path.glob because you mgiht pass
                # 'reqs/*.txt' and this is considered a non-relative pattern.
                yield from iter_requirement_names(Path(filename))



import functools
import logging
from typing import Dict, List, Optional, Tuple

from keke import kev

from packaging.requirements import Requirement
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import Version
from vmodule import VLOG_1, VLOG_2

from .markers import EnvironmentMarkers
from .projects import Project
from .types import VersionCallback

LOG = logging.getLogger(__name__)


class NoMatchingRelease(Exception):
    pass


def requires_python_match(
    project: Project, cache: Dict[str, bool], python_version: Version, v: Version
) -> bool:
    pv = project.versions[v]
    if not pv.requires_python:
        return True
    if (verdict := cache.get(pv.requires_python)) is not None:
        return verdict
    try:
        specifier_set = SpecifierSet(pv.requires_python)
        verdict = python_version in specifier_set
    except InvalidSpecifier as e:
        LOG.debug(
            "Ignore %s==%s has invalid requires_python %r but including anyway",
            project.name,
            v,
            e,
        )
        verdict = True
    cache[pv.requires_python] = verdict
    return verdict


def find_best_compatible_version(
    project: Project,
    req: Requirement,
    env_markers: EnvironmentMarkers,
    already_chosen: Optional[Version],
    current_version_callback: VersionCallback,
) -> Version:
    # Handle requires_python first, so we can produce a better error message
    # when there are no version-compatible candidates (before we even get to
    # req.specifier)
    python_version_str = env_markers.python_full_version
    assert python_version_str is not None
    python_version = Version(python_version_str)

    _requires_python_match = functools.partial(
        requires_python_match, project, {}, python_version
    )

    possible: List[Version] = []

    specifier_matched = False
    with kev("reverse"):
        rev = reversed(project.versions.keys())
    with kev("initial filter"):
        for version in req.specifier.filter(rev):
            specifier_matched = True
            if _requires_python_match(version):
                # Only keep the first one
                possible.append(version)
                break

    # If the current version is a non-public version, we need to add it back
    # here.
    with kev("current_version_callback"):
        cur = current_version_callback(project.name)
    cur_v: Optional[Version] = None
    if cur:
        cur_v = Version(cur)
        # Allow the current version to be a guess -- it doesn't actually have to
        # be compatible with the current version of python.  Filter out if we
        # know for sure.
        if cur_v in project.versions:
            if _requires_python_match(cur_v):
                possible.append(cur_v)
        else:
            # Non-public version, assume it's ok
            possible.append(cur_v)
    LOG.log(VLOG_2, "Callback %s -> %s", cur, cur_v)

    if already_chosen:
        possible.append(already_chosen)

    if not possible:
        if specifier_matched:
            raise NoMatchingRelease(
                f"{project.name} has no {python_version}-compatible release"
            )
        else:
            raise NoMatchingRelease(
                f"{project.name} has no release matching {req.specifier}"
            )

    # The documentation for SepcifierSet.filter notes that it handles the logic
    # for whether to include prereleases, so we don't need that here.
    LOG.log(VLOG_1, "possible for %s: %s", req, possible)
    with kev("filter by specifier"):
        # This should only ever be ~3 items now!
        possible = list(req.specifier.filter(possible))
    if not possible:
        if not list(req.specifier.filter(project.versions.keys())):
            # Referencing the dragon above, if we had a current version and it was
            # unsuitable, then we still output a generic message.  Note that > does not
            # set the prerelease bit, but >= does.
            raise NoMatchingRelease(
                f"{project.name} has no release with constraint {req.specifier}"
            )
        else:
            raise NoMatchingRelease(
                f"{project.name} has no {python_version}-compatible release with constraint {req.specifier}"
            )

    # TODO: yanked support

    # The sort key is
    # ( matches_already_chosen: bool,
    #   matches_current_version: bool,
    #   recency_index: int,
    #   version: Version )
    # so that after sorting the last item is the "best" one.

    with kev("final sort"):
        xform_possible: List[Tuple[bool, bool, int, Version]] = sorted(
            (p == already_chosen, p == cur_v, i, p) for (i, p) in enumerate(possible)
        )

    LOG.log(VLOG_1, "Sorted %s", project.name)
    for el in xform_possible:
        LOG.log(VLOG_1, "%s", el)
    return xform_possible[-1][3]
